/*********************************************************************
Copyright 1992 by Digital Press/Digital Equipment Corporation, Elias
Israel, and Erik Fortune

Permission to use, copy, modify, distribute, and sell this software
and its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation.  Digital Equipment Corporation and the
authors make no representations about the suitability of this software
for any purpose.  It is provided "as is" without express or implied
warranty.

DIGITAL EQUIPMENT CORPORATION AND THE AUTHORS DISCLAIM ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL DIGITAL EQUIPMENT
CORPORATION OR THE AUTHORS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

This material is reproduced with permission from _The X Window System
Server: X Version 11, Release 5_, by Elias Israel and Erik Fortune.
(Copyright 1992, Digital Press, Digital Equipment Corporation, One
Burlington Woods Drive, Burlington, MA 01803)
*********************************************************************/

#include "X.h"
#define NEED_EVENTS
#include "Xproto.h"
#include "misc.h"
#include "dixstruct.h"
#include "inputstr.h"

	/*
	 *     The input extension allows implementors to enable all devices 
	 * when the server starts up, but we have found it best to close 
	 * devices when no clients are using them.  Unfortunately, it can be 
	 * difficult to determine when to enable or disable input devices 
	 * other than the core pointer and keyboard device (which are always 
	 * open).    It is not enough to simply count the number of times a 
	 * device is opened and closed for two reasons:
	 *     o Some clients exit without explicitly closing the device.
	 *     o Clients might open a device several times but only close
	 *       it once.   The input extension specifies that a call to
	 *       CloseDevice from a client closes the device for that client
	 *       regardless of the number of times that the client has opened
	 *       it.   Of course, we don't want to disable the device if some
	 *       other client is still using it.
	 *     To (mostly) solve the first problem, we can use a resource to
	 * automatically close the device when the client exits -- this won't
	 * work if the client exits with a close-down mode other than 
	 * DestroyAll, but this doesn't cause problems in practice.
	 *     To solve the second problem, we can use the devicePrivate of
	 * the input device to keep track of which clients have opened the
	 * device.
	 *     This file partially implements these solutions.  In order to 
	 * use these functions, you must:
	 *     o Call ddxInitInputExtension from InitInput once per server 
	 *       generation.
	 *     o Modify InitInput (or ddxInitInputExtension) to locate,
	 *       initialize and register all extension devices.
	 *     o Write a device procedure for extension devices and modify 
	 *       the PointerProc and KeyboardProc in ../xos to allocate
	 *       devicePrivates with the necessary information.
	 *     o Implement a number of other DDX-layer functions used by
	 *       the input extension.  See the file
	 *       mit/extensions/server/xinput/xstubs.c for a description of
	 *       each of these functions.
	 */

RESTYPE	RT_INPUTCLIENT;

typedef struct _ddxInputPriv {
	int	isCore;
	int	nOpen;	   /* number of clients currently using the device */
	int	szClients; /* number of entries in the "clients" array */
	struct ddxInputClient {
	    int	    index;    /* the index of this client */
	    XID     resource; /* the resource used to close on exit */
	} *clients;
} ddxInputPriv;

	/*
	 *    This routine returns a resource ID for pClient to use pXDev.   
	 *    If the specified client is already using this device, simply 
	 * return the old resource ID.   
	 *    If the client is not using this device, allocate and return
	 * a new ID for it (after storing it in the devicePrivate for
	 * the device).   If the device is not currently open, call
	 * EnableDevice to turn it on.
	 */

static XID
AddClientResource(pXDev,pClient)
DeviceRec *pXDev;
ClientPtr  pClient;
{
ddxInputPriv	*pPriv=	(ddxInputPriv *)pXDev->devicePrivate;
int		 i;
XID		 id;

    for (i=0;i<pPriv->nOpen;i++) {
	if (pPriv->clients[i].index==pClient->index) {
	    return pPriv->clients[i].resource;
	}
    }
    if (pPriv->nOpen>=pPriv->szClients-1) {
	if (pPriv->szClients==0) {
	    pPriv->szClients= 1;
	    pPriv->clients= (struct ddxInputClient *)
					xalloc(sizeof(struct ddxInputClient));
	}
	else {
	    pPriv->szClients*= 2;
	    pPriv->clients= (struct ddxInputClient *)
			xrealloc(pPriv->clients,
				pPriv->szClients*sizeof(struct ddxInputClient));
	}
	if (pPriv->clients==NULL) {
	    ErrorF("Error! Allocation failed in AddClientResource\n");
	    pPriv->nOpen=	0;
	    DisableDevice(pXDev);
	    return 0;
	}
    }
    id= FakeClientID(pClient->index);
    AddResource(id,RT_INPUTCLIENT,pXDev);
    pPriv->clients[pPriv->nOpen].index=	pClient->index;
    pPriv->clients[pPriv->nOpen].resource=	id;
    pPriv->nOpen++;
    if (!pXDev->on)
	EnableDevice(pXDev);
    return id;
}

    /*
     *    Given a device and a client index, it finds the resource
     * ID that was created when the client opened the device.   If
     * the client is not using the device, FindClientResource returns
     * INVALID.
     */
static XID
FindClientResource(pXDev,index)
DeviceRec *pXDev;
int	   index;
{
ddxInputPriv	*pPriv=	(ddxInputPriv*)pXDev->devicePrivate;
int		 i;

    for (i=0;i<pPriv->nOpen;i++) {
	if (pPriv->clients[i].index==index) {
	    return pPriv->clients[i].resource;
	}
    }
    return INVALID;
}

	/*
	 * Given a device and a resource ID, remove the client which
	 * corresponds to the resource ID from the devicePrivate
	 * of pXDev.   Disable the device if it is not a core device
	 * and no more clients are using it.
	 */
static int
RemoveResourceClient(pXDev,resource)
DeviceRec *pXDev;
XID	   resource;
{
ddxInputPriv	*pPriv=	(ddxInputPriv *)pXDev->devicePrivate;
register int	 i;

    for (i=0; i < pPriv->nOpen; i++) {
	if (pPriv->clients[i].resource == resource) {
	    pPriv->nOpen--;
	    for (; i < pPriv->nOpen; i++) {
		pPriv->clients[i]= pPriv->clients[i+1];
	    }
	    if ((pPriv->nOpen<1) && (!pPriv->isCore))
		DisableDevice( pXDev );
	    return 1;
	}
    }
    return 0;
}

	/*
	 * The Input Extension calls this routine whenever a client
	 * opens an extension device, whether or not that client is
	 * already using the device.   Try to create a resource for
	 * the client and return Success if we get one.
	 */
void
OpenInputDevice(pXDev,pClient,status)
DeviceIntRec *pXDev;
ClientRec    *pClient;
int          *status;
{
XID	id;

    id = AddClientResource(pXDev,pClient,id);
    if (id!=INVALID) {
	*status= Success;
	return;
    }
    /* 
     * This can only happen on an internal error or if we run out
     * of resources (unlikely).   We don't really know why it failed,
     * but BadAlloc is as good a guess as any.   BadImplementation
     * is another sterling choice.
     */
    *status=	BadAlloc;
    return;
}

    /*
     * The Input Extension calls this routine whenever a client closes
     * an input device.   It should remove the client from the 
     * devicePrivate list of clients using the device.
     */

void
CloseInputDevice(pXDev,pClient)
DeviceIntRec *pXDev;
ClientRec    *pClient;
{
XID	id= FindClientResource(pXDev,pClient->index);

    if (id!=INVALID) {
	/* 
	 * This call to FreeResource will call ddxIEClientGone (below)
	 * which takes care of actually cleaning up the devicePrivates.
	 * We don't have to do anything further here.
	 */
	FreeResource(id,RT_NONE);
	return;
    }
    ErrorF("Internal Error! No client resource in CloseInputDevice\n");
    return;
}

	/*
	 * The DIX layer resource code calls this routine whenever
	 * a resource of type RT_INPUTCLIENT is freed.   It calls
	 * RemoveResourceClient to clean up the devicePrivates and
	 * close the device if necessary.
	 */
static int
ddxIEClientGone(data,id)
pointer data;
XID	id;
{
DeviceRec	*pXDev=	(DeviceRec *)data;

    if (!RemoveResourceClient(pXDev,id)) {
	ErrorF("Internal Error! RemoveResource failed in ClientGone\n");
    }
    return 1;
}

	/*
	 *     We have to create the new resource type here.   If you
	 * use these device privates, be sure to call this function
	 * from InitInput.    Call it once every server generation.
	 */
void
ddxInitInputExtension()
{
    RT_INPUTCLIENT=	CreateNewResourceType(ddxIEClientGone);
    return;
}

	/*
	 * The following two functions make it possible to remove
	 * input devices if they disappear and are described in chapter 
	 * 15.
	 */

extern InputInfo inputInfo;
static int highID = 0;

int
FindDeviceID()
{
    if (highID<MAX_DEVICES)
	 return highID++;
    else {
	char inuse[MAX_DEVICES/sizeof(char)+1];
	register int i;
	register DeviceIntPtr dev;

	bzero(inuse,sizeof(inuse));
	dev= inputInfo.off_devices;
	while (dev) {
	    inuse[dev->id/8]|= (1<<(dev->id%8));
	    dev= dev->next;
	}
	dev= inputInfo.devices;
	while (dev) {
	    inuse[dev->id/8]|= (1<<(dev->id%8));
	    dev= dev->next;
	}
	for (i=0;i<MAX_DEVICES;i++) {
	    if (!(inuse[i/8]&(1<<(i%8))))
		return i;
	}
	ErrorF("Warning! Couldn't find a device ID.\n");
	return MAX_DEVICES;
    }
}

void
RemoveDevice(dev)
    register DeviceIntPtr dev;
{
    register DeviceIntPtr prev,tmp,next;

    prev= NULL;
    for (tmp= inputInfo.devices; tmp; (prev = tmp), (tmp = next)) {
	next = tmp->next;
	if (tmp==dev) {
	    CloseDevice(tmp);
	    if (prev==NULL)
		inputInfo.devices = next;
	    else
	        prev->next = next;
	    inputInfo.numDevices--;
	    return;
	}
    }

    prev= NULL;
    for (tmp= inputInfo.off_devices; tmp; (prev = tmp), (tmp = next)) {
	next = tmp->next;
	if (tmp==dev) {
	    CloseDevice(tmp);
	    if (prev==NULL)
		inputInfo.off_devices = next;
	    else
		prev->next = next;
	    inputInfo.numDevices--;
	    return;
	}
    }
    ErrorF("Internal Error! Attempt to remove a non-existent device\n");
    return;
}
